#include <libregexis024fa/codeset.h>
#include <assert.h>

namespace regexis024 {
    codeset_t invert_set(const codeset_t &X) {
        if (X.empty())
            return {{0, UINT32_MAX}};
        codeset_t res;
        if (X[0].first != 0)
            res.emplace_back(0, X[0].first - 1);
        for (size_t i = 0; i + 1 < X.size(); i++){
            res.emplace_back(X[i].second + 1, X[i + 1].first - 1);
        }
        if (X.back().second != UINT32_MAX)
            res.emplace_back(X.back().second + 1, UINT32_MAX);
        return res;
    }

#define elA (A[i])
#define elB (B[j])
#define Ainc i++
#define Binc j++
#define prepare size_t An = A.size(); size_t Bn = B.size(); size_t i = 0; size_t j = 0;
#define Aended (i == An)
#define Bended (j == Bn)

    codeset_t merge_sets(const codeset_t &A, const codeset_t &B) {
        codeset_t res;
        prepare
        std::pair<uint32_t, uint32_t> cur;
        while (true){
            if (Aended && Bended)
                break;
            if (i == An){
                cur = elB;
                Binc;
            } else if (j == Bn){
                cur = elA;
                Ainc;
            } else {
                if (elA.first < elB.first) {
                    cur = elA;
                    Ainc;
                } else {
                    cur = elB;
                    Binc;
                }
            }
            while (true){
                if (Aended && Bended){
                    res.push_back(cur);
                    break;
                }
                if (i < An && (cur.second == UINT32_MAX || elA.first <= cur.second + 1)){
                    cur.second = std::max(elA.second, cur.second);
                    Ainc;
                } else if (j < Bn && (cur.second == UINT32_MAX || elB.first <= cur.second + 1)){
                    cur.second = std::max(elB.second, cur.second);
                    Binc;
                } else {
                    res.push_back(cur);
                    break;
                }
            }
        }
        return res;
    }

    codeset_t intersect_sets(const codeset_t &A, const codeset_t &B) {
        codeset_t res;
        prepare
        while (true){
            if (Aended || Bended)
                break;
            if (elB.first <= elA.first && elA.first <= elB.second)
                res.emplace_back(elA.first, std::min(elA.second, elB.second));
            else if (elA.first <= elB.first && elB.first <= elA.second)
                res.emplace_back(elB.first, std::min(elA.second, elB.second));

            if (elA.second <= elB.second)
                Ainc;
            else
                Binc;
        }
        return res;
    }

    codeset_t subtract_sets(const codeset_t &A, const codeset_t &B) {
        return intersect_sets(A, invert_set(B));
    }

    bool is_inside(uint32_t start, uint32_t end, codeset_t &X) {
        for (auto& p: X){
            if (p.first <= start && end <= p.second)
                return true;
            assert(end < p.first || p.second < start);
        }
        return false;
    }

    codeset_t set_add_char(const codeset_t& X, uint32_t cp) {
        return merge_sets(X, {{cp, cp}});
    }

    codeset_t set_add_range(const codeset_t& X, uint32_t start, uint32_t end) {
        return merge_sets(X, {{start, end}});
    }

    codeset_t codeset_of_one_char(uint32_t ch) {
        return codeset_t({{ch, ch}});
    }

    std::string stringifyCodesetBase10(const codeset_t& CS) {
        std::string cs;
        for (auto p: CS) {
            if (!cs.empty())
                cs += "; ";
            cs += std::to_string(p.first) + "-" + std::to_string(p.second);
        }
        return cs;
    }
}
