#include <libregexis024fa/graph_to_bytecode/fa_compiler.h>

#include <libregexis024vm/vm_opcodes.h>
#include <libregexis024fa/graph_to_bytecode/natural_compiler_utils.h>
#include <assert.h>
#include <libregexis024fa/graph_to_bytecode/writing_commands.h>

#include <libregexis024fa/graph_to_bytecode/core.h>

namespace regexis024 {
    void write_priority_table_actions(std::vector<uint8_t>& result, RegexPriorityTable &priority_table) {
        for (RegexPriorityTableAction& act: priority_table) {
            if (act.pos.isForRange()) {
                write_byte(result, opcodes::DDIST_RABX_SELARR);
                write_tai(result, act.pos.first);
                write_tai(result, act.pos.second);
            } else {
                write_byte(result, opcodes::DMOV_RABX_SELARR);
                write_tai(result, act.pos.first);
            }
            write_byte(result, act.minimize ?
                opcodes::SIFTPRIOR_MIN_RABX :
                opcodes::SIFTPRIOR_MAX_RABX);
        }
        write_byte(result, opcodes::SIFT_DONE);
    }

    struct belate_initialization_parameters {
        size_t todo_pos_read_ss_n;
        size_t todo_pos_fork_ss_n;

        void complete_it(std::vector<uint8_t>& result, sslot_id_t read_ss_ns, sslot_id_t fork_ss_ns)
        {
            belated_sslot_id(result, todo_pos_read_ss_n , read_ss_ns);
            belated_sslot_id(result, todo_pos_fork_ss_n, fork_ss_ns);
        }
    };

    /* when I compile initializational part of program, I don't yet know what to put in
     * PARAM_READ_SS_NUMBER, PARAM_FORK_SS_NUMBER and MSG_FED_INPUT_EXTENDED (second namespace size).
     * These values are belate. */
    belate_initialization_parameters write_some_normal_initialization(std::vector<uint8_t>& result,
        size_t selarr_size, const REGEX_IS024_FA_FirstStageFixInfo& info1)
    {
        belate_initialization_parameters todo;

        write_byte(result, opcodes::PARAM_READ_SS_NUMBER);
        todo.todo_pos_read_ss_n = result.size();
        write_sslot_id(result, 0);  // Belate

        write_byte(result, opcodes::PARAM_FORK_SS_NUMBER);
        todo.todo_pos_fork_ss_n = result.size();
        write_sslot_id(result, 0);  // Belate

        write_byte(result, opcodes::PARAM_SELARR_LEN);
        write_tai(result, selarr_size);

        write_byte(result, opcodes::INIT);
        return todo;
    }

    void compile_fa_to_regexis024_bytecode(std::vector<uint8_t>& result,
        FA_Container &fa, RegexPriorityTable &priority_table,
        size_t selarr_size, const REGEX_IS024_FA_FirstStageFixInfo& info1, int& error)
    {
        error = 0;
        explicit_bookmarks bookmark_manager;

        if (!priority_table.empty()) {
            bookmark_id_t BM_sift_function = bookmark_manager.new_bookmark();
            bookmark_id_t BM_after_sift = bookmark_manager.new_bookmark();

            cmd_JUMP(result, bookmark_manager, BM_after_sift);
            bookmark_manager.land_bookmark(result, BM_sift_function);
            write_priority_table_actions(result, priority_table);
            bookmark_manager.land_bookmark(result, BM_after_sift);

            write_byte(result, opcodes::PARAM_COLSIFTFUNC_SET);
            bookmark_manager.write_unresolved_reference(result, BM_sift_function);
        }

        belate_initialization_parameters init_param_todo = write_some_normal_initialization(result, selarr_size, info1);

        size_t read_ss_ns, fork_ss_ns;
        compilation_core(result, fa, bookmark_manager, read_ss_ns, fork_ss_ns, error);
        if (error < 0)
            return;
        init_param_todo.complete_it(result, read_ss_ns, fork_ss_ns);
        bookmark_manager.finish(result);
    }
}
