#include <libregexis024fa/graph_to_bytecode/natural_compiler_utils.h>
#include <assert.h>
#include <libregexis024vm/vm_opcodes.h>

namespace regexis024 {
#define push_to_res_least_signif result.push_back(x & 0xffLU); x >>= 8
    void write_byte(std::vector<uint8_t>& result, uint8_t x) {
        result.push_back(x);
    }

    void write_word(std::vector<uint8_t>& result, uint16_t x) {
        push_to_res_least_signif; push_to_res_least_signif;
    }

    void write_doubleword(std::vector<uint8_t>& result, uint32_t x) {
        push_to_res_least_signif; push_to_res_least_signif; push_to_res_least_signif; push_to_res_least_signif;
    }

    void write_quadword(std::vector<uint8_t>& result, uint64_t x) {
        for (int i = 0; i < 8; i++) {
            push_to_res_least_signif;
        }
    }
#undef push_to_res_least_signif

#define put_belated_to_res assert(result[pos] == 0); result[pos++] = value & 0xffLU; value >>= 8
    void belated_byte(std::vector<uint8_t>& result, size_t pos, uint8_t value) {
        assert(pos < result.size());
        result[pos] = value;
    }

    void belated_word(std::vector<uint8_t>& result, size_t pos, uint16_t value) {
        assert(pos + 2 <= result.size());
        put_belated_to_res; put_belated_to_res;
    }

    void belated_doubleword(std::vector<uint8_t>& result, size_t pos, uint32_t value) {
        assert(pos + 4 <= result.size());
        put_belated_to_res; put_belated_to_res; put_belated_to_res; put_belated_to_res;
    }

    void belated_quadword(std::vector<uint8_t>& result, size_t pos, uint64_t value) {
        assert(pos + 8 <= result.size());
        for (int i = 0; i < 8; i++) {
            put_belated_to_res;
        }
    }
#undef put_belated_to_res

    void write_sslot_id(std::vector<uint8_t>& result, sslot_id_t x) {
        write_doubleword(result, x);
    }

    void write_tai(std::vector<uint8_t>& result, tai_t x) {
        write_word(result, x);
    }

    void write_near_ptr(std::vector<uint8_t>& result, near_ptr_t x) {
        write_quadword(result, x);
    }

    void belated_sslot_id(std::vector<uint8_t>& result, size_t pos, sslot_id_t value) {
        belated_doubleword(result, pos, value);
    }

    void belated_tai(std::vector<uint8_t>& result, size_t pos, tai_t value) {
        belated_word(result, pos, value);
    }

    void belated_near_ptr(std::vector<uint8_t>& result, size_t pos, near_ptr_t value) {
        belated_quadword(result, pos, value);
    }

    bookmark_id_t explicit_bookmarks::new_bookmark() {
        pile.emplace_back();
        return free_bid++;
    }

    void explicit_bookmarks::write_unresolved_reference(std::vector<uint8_t> &result, bookmark_id_t bm) {
        size_t where_to_fill_later = result.size();
        write_near_ptr(result, 0);
        pile[bm].positions_of_belated_refs.push_back(where_to_fill_later);
    }

    void explicit_bookmarks::land_bookmark(std::vector<uint8_t> &result, bookmark_id_t bm) {
        assert(!pile[bm].placed_somewhere);
        pile[bm].placed_somewhere = true;
        pile[bm].actual_position = result.size();
    }

    void explicit_bookmarks::finish(std::vector<uint8_t> &result) {
        for (explicit_bookmark_info& bmi: pile) {
            assert(bmi.positions_of_belated_refs.empty() || bmi.placed_somewhere);
            if (bmi.placed_somewhere) {
                for (size_t ref_to_mine_belate: bmi.positions_of_belated_refs) {
                    belated_near_ptr(result, ref_to_mine_belate, bmi.actual_position);
                }
            }
        }
    }

    bookmark_id_t explicit_bookmarks::new_range_of_bookmarks(size_t n) {
        bookmark_id_t offset = free_bid;
        free_bid += n;
        for (size_t i = 0; i < n; i++) {
            pile.emplace_back();
        }
        return offset;
    }

    bool explicit_bookmarks::has_landed(bookmark_id_t bm) {
        return pile[bm].placed_somewhere;
    }
}

#undef put_belated_to_res
