#ifndef LIBREGEXIS024_SRC_LIBREGEXIS024FA_NATURAL_COMPILER_UTILS_H
#define LIBREGEXIS024_SRC_LIBREGEXIS024FA_NATURAL_COMPILER_UTILS_H

#include <stdint.h>
#include <libregexis024vm/vm_opcodes_types.h>
#include <vector>
#include <stddef.h>

namespace regexis024 {
    void write_byte(std::vector<uint8_t>& result, uint8_t x);
    void write_word(std::vector<uint8_t>& result, uint16_t x);
    void write_doubleword(std::vector<uint8_t>& result, uint32_t x);
    void write_quadword(std::vector<uint8_t>& result, uint64_t x);

    void belated_byte(std::vector<uint8_t>& result, size_t pos, uint8_t value);
    void belated_word(std::vector<uint8_t>& result, size_t pos, uint16_t value);
    void belated_doubleword(std::vector<uint8_t>& result, size_t pos, uint32_t value);
    void belated_quadword(std::vector<uint8_t>& result, size_t pos, uint64_t value);


    void write_sslot_id(std::vector<uint8_t>& result, sslot_id_t x);
    void write_tai(std::vector<uint8_t>& result, tai_t x);
    void write_near_ptr(std::vector<uint8_t>& result, near_ptr_t x);

    void belated_sslot_id(std::vector<uint8_t>& result, size_t pos, sslot_id_t value);
    void belated_tai(std::vector<uint8_t>& result, size_t pos, tai_t value);
    void belated_near_ptr(std::vector<uint8_t>& result, size_t pos, near_ptr_t value);

    // constexpr uint64_t INSTRUCTION_SZ = REGEX024_BYTECODE_INSTRUCTION_SZ;
    // constexpr uint64_t SSLOT_ID_SZ = REGEX024_BYTECODE_SSLOT_ID_SZ;
    // constexpr uint64_t TRACK_ARRAY_INDEX_ID_SZ = REGEX024_BYTECODE_TRACK_ARRAY_INDEX_ID_SZ;
    // constexpr uint64_t NEAR_POINTER_SZ = REGEX024_BYTECODE_NEAR_POINTER_SZ;

    typedef size_t bookmark_id_t;

    struct explicit_bookmark_info {
        std::vector<size_t> positions_of_belated_refs;
        bool placed_somewhere = false;
        size_t actual_position;
    };

    struct explicit_bookmarks {
        bookmark_id_t free_bid = 0;
        /* For each named explicit bookmark there is an element in PILE */
        std::vector<explicit_bookmark_info> pile;

        bookmark_id_t new_bookmark();

        /* bm is the bookmark I refer to. Each bookmark has an id. It is like a name, but fits in 8 bytes */
        void write_unresolved_reference(std::vector<uint8_t>& result,  bookmark_id_t bm);

        /* bm is the bookmark I place into program `result` */
        void land_bookmark(std::vector<uint8_t>& result, bookmark_id_t bm);

        /* call it at the very end of bytecode-building */
        void finish(std::vector<uint8_t>& result);

        /* Returns offset of range of bookmark id's */
        bookmark_id_t new_range_of_bookmarks(size_t n);

        bool has_landed(bookmark_id_t bm);
    };
}

#endif
