#include <libregexis024fa/graph_to_bytecode/writing_commands.h>
#include <libregexis024vm/vm_opcodes.h>
#include <assert.h>

namespace regexis024 {
    void cmd_JUMP(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, bookmark_id_t dest) {
        write_byte(result, opcodes::JUMP);
        bookmark_manager.write_unresolved_reference(result, dest);
    }

    constexpr opcode_t cmp_EQUAL[4] = {opcodes::JCEQUAL_B, opcodes::JCEQUAL_W,
        opcodes::JCEQUAL_DW, opcodes::JCEQUAL_QW};
    constexpr opcode_t cmp_LESS[4] = {opcodes::JCLESS_B, opcodes::JCLESS_W,
        opcodes::JCLESS_DW, opcodes::JCLESS_QW};
    constexpr opcode_t cmp_GRTR[4] = {opcodes::JCGRTR_B, opcodes::JCGRTR_W,
        opcodes::JCGRTR_DW, opcodes::JCGRTR_QW};


    void cmd_JC(const opcode_t cmpT[4],
        std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest)
    {
        if (val <= UINT8_MAX) {
            write_byte(result, cmpT[0]);
            write_byte(result, static_cast<uint8_t>(val));
        } else if (val <= UINT16_MAX) {
            write_byte(result, cmpT[1]);
            write_word(result, static_cast<uint16_t>(val));
        } else if (val <= UINT32_MAX) {
            write_byte(result, cmpT[2]);
            write_doubleword(result, static_cast<uint32_t>(val));
        } else {
            write_byte(result, cmpT[3]);
            write_quadword(result, val);
        }
        bookmark_manager.write_unresolved_reference(result, dest);
    }


    void cmd_JCEQUAL(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest) {
        cmd_JC(cmp_EQUAL, result, bookmark_manager, val, dest);
    }

    void cmd_JCLESS(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest) {
        cmd_JC(cmp_LESS, result, bookmark_manager, val, dest);
    }

    void cmd_JCGRTR(std::vector<uint8_t>& result, explicit_bookmarks& bookmark_manager, uint64_t val, bookmark_id_t dest) {
        cmd_JC(cmp_GRTR, result, bookmark_manager, val, dest);
    }

    void cmd_DIE(std::vector<uint8_t> &result) {
        write_byte(result, opcodes::DIE);
    }

    void cmd_MATCH(std::vector<uint8_t> &result) {
        write_byte(result, opcodes::MATCH);
    }

    void cmd_READ(std::vector<uint8_t>& result, size_t slot) {
        assert(slot <= UINT32_MAX);
        write_byte(result, opcodes::READ);
        write_sslot_id(result, slot);
    }

    void cmd_FORK(std::vector<uint8_t> &result, explicit_bookmarks& bookmark_manager, size_t slot, bookmark_id_t dest) {
        assert(slot <= UINT32_MAX);
        write_byte(result, opcodes::FORK);
        write_sslot_id(result, slot);
        bookmark_manager.write_unresolved_reference(result, dest);
    }
}