#include <stdio.h>

#include <libregexis024vm/libregexis024vm.h>
#include <libregexis024vm/vm_opcodes.h>
#include <libregexis024test/byte_code_assembler.h>
#include <libregexis024test/byte_code_disassembler.h>
#include <assert.h>

static int test_id = 0;

void do_test(const std::vector<uint8_t>& prg, const std::string& str, const std::vector<bool>& prefix_matching){
    assert(str.size() + 1 == prefix_matching.size());
    VMContext ctx{prg.size(), prg.data(), 0, 0, 1000, 1000, 1000000};
    error_code_t ret;
    // todo
    printf("TEST %d passed\n", test_id);
    test_id++;
}

std::vector<bool> tfff(size_t n){
    std::vector<bool> res = std::vector<bool>(n + 1, false);
    res[0] = true;
    return res;
}

std::vector<bool> ffft(size_t n){
    std::vector<bool> res = std::vector<bool>(n + 1, false);
    res.back() = true;
    return res;
}

std::vector<uint8_t> program_0(){
    std::vector<uint8_t> res;
    s_BEGIN_ASSEMBLER_CONTEXT()
        i_PARAM_READ_SS_NUMBER(2)
        i_PARAM_FORK_SS_NUMBER(1)
        i_INIT()
        i_FORK(0, "wb")
        i_READ(0)
        i_JCEQUAL_B('a', "finish")
        i_DIE()

        c_BOOKMARK("wb")
        i_READ(1)
        i_JCEQUAL_B('b', "finish")
        i_DIE()

        c_BOOKMARK("finish")
        i_MATCH()
        i_DIE()
    s_END_ASSEMBLER_CONTEXT(res)
    return res;
}

std::vector<uint8_t> program_1(){
    std::vector<uint8_t> res;
    s_BEGIN_ASSEMBLER_CONTEXT()
        i_PARAM_READ_SS_NUMBER(4)
        i_PARAM_FORK_SS_NUMBER(2)
//        i_PARAM_SELARR_LEN(0)
        i_INIT()
        i_FORK(0, "wb")
        c_BOOKMARK("wa")
        i_READ(0)
        i_JCEQUAL_B('a', "razvilka")
        i_DIE()

        c_BOOKMARK("wb")
        i_READ(1)
        i_JCEQUAL_B('b', "razvilka")
        i_DIE()

        c_BOOKMARK("razvilka")
        i_FORK(1, "wd")
        c_BOOKMARK("wc")
        i_READ(2)
        i_JCEQUAL_B('c', "finish")
        i_DIE()

        c_BOOKMARK("wd")
        i_READ(3)
        i_JCEQUAL_B('d', "finish")
        i_DIE()

        c_BOOKMARK("finish")
        i_MATCH()
        i_DIE()
    s_END_ASSEMBLER_CONTEXT(res)
    return res;
}

/*
int main(){
    auto prg0 = program_0();
    auto prg1 = program_1();
//    printf("Disassembled program:\n");
//    print_disassembly(prg.size(), prg.data());

    printf("Testing starts\n");
    test_id = 0;
    do_test(prg0, "a", {false, true});
    do_test(prg0, "b", {false, true});
    do_test(prg0, "c", {false, false});
    do_test(prg0, "a4", {false, true, false});
    do_test(prg0, "b4", {false, true, false});
    do_test(prg1, "aa", {false, false, false});
    do_test(prg1, "db", {false, false, false});
    do_test(prg1, "ac", {false, false, true});
    do_test(prg1, "bc", {false, false, true});
    do_test(prg1, "ad", {false, false, true});
    do_test(prg1, "bd", {false, false, true});
    do_test(prg1, "bd12", {false, false, true, false, false});

    return 0;
}
*/

int main() {
    std::vector<uint8_t> prg;
    s_BEGIN_ASSEMBLER_CONTEXT()
        c_BOOKMARK("111")
        i_READ(0)
        i_READ(1)
        i_FORK(12, "vdv")
        c_BOOKMARK("vdv")
        i_READ(2)
        i_READ(10000000)
        i_THROW()
        i_THROW()
        i_THROW()
        i_JUMP("111")

    s_END_ASSEMBLER_CONTEXT(prg)

    print_disassembly(prg.size(), prg.data());
}