#include <libregexis024tools/stringmatching.h>
#include <assert.h>
#include <stdexcept>
#include <libregexis024sol/expr_compiler.h>
#include <libregexis024tools/delayed_matching.h>

using namespace regexis024;
using namespace std;

void test(const string& input, const string& pattern, const MatchInfo& right_answer) {
    MatchInfo given_answer;
    track_var_list retTrackVarList;
    string retStatus;
    int ret = matchStrToRegexp(input, pattern, given_answer, retTrackVarList, retStatus);
    if (ret < 0) {
        throw runtime_error("Test failed. matching. " + retStatus);
    }
    if (given_answer != right_answer) {
        throw runtime_error("Test failed.");
    }
    printf("Test passed\n");
}

void boba(const vector<pair<string, bool>>& input, const string& pattern) {
    std::vector<uint8_t> regexp_program;
    std::map<std::string, TrackingVariableInfo> variables;
    std::string error;
    int ret = compile(pattern, variables, regexp_program, error);
    if (ret < 0) {
        throw runtime_error("Compilation failure " + error);
    }
    VirtualMachine vm(regexp_program.size(), regexp_program.data(),
            UINT64_MAX, UINT16_MAX,
            UINT32_MAX, UINT32_MAX, UINT64_MAX);
    if (vm.initialize() < 0) {
        throw runtime_error("Can't initialize");
    }
    for (int i = 0;  i < 3; i++) {
        for (auto& inp: input) {
            vm.wipeToInit();
            if (vm.addNewMatchingThread() < 0)
                throw runtime_error("addNewMatchingThread");
            for (char ch: inp.first) {
                if (vm.feedCharacter(ch, 1) < 0)
                    throw runtime_error("feedCharacter");
            }
            if (vm.isMatched() != inp.second)
                throw runtime_error("AAAAAAAAAAAAAAAA");
        }
    }
}

int main() {
#define reg_ALPHA "[a-zA-Z]"
#define reg_pchar "([a-zA-Z0-9\\-._~\\!$\\&'()*+@:,;=]|%[0-9a-hA-H]!r{2})"
#define reg_query "(" reg_pchar"|[?/])*"
#define reg_lin_ws "([ \t]|\r\n[ \t])*"
#define reg_request_line "#method(" reg_ALPHA"+) #uri_path(/(" reg_pchar"|/)*)(\\?#uri_query(" reg_query"))? HTTP/#http_version(!digit;+.!digit;+)\r\n"
#define reg_filed_value "(" reg_lin_ws"#header_field_value_part([\\u0021-\\u007e&^\r\n]+))*" reg_lin_ws
#define reg_HTTP_message reg_request_line "(#header_field_name([\\u0021-\\u007E&^:]+):" reg_filed_value "\r\n)*\r\n"
    test("P /aa?:: HTTP/1.1\r\nr--r: qwe\r\n\r\n", reg_HTTP_message, MatchInfo(
        {{0, 0}, {1, 1}, {2, 2}, {3, 5}, {4, 6}, {5, 8}, {6, 14}, {7, 17}, {8, 19}, {9, 23}, {10, 25}, {11, 28}}, {}));
    test("d3", "[abc]3", MatchInfo());

    boba({{"", true}, {"a", false}}, "");
    boba({{"", true}, {"a", true}, {"aab", true}, {"aac", false}, {"c", false}, {"abbababa", true}}, "[ab]*");
    test("", "", MatchInfo({}, {}));
    test("a", "a", MatchInfo({}, {}));
    test("1b", "1\\<b", MatchInfo({}, {}));
    test("bababbaa", "[ab]*", MatchInfo({}, {}));
    test("bababbaa", "!dfa;[ab]*", MatchInfo({}, {}));
    test("a3", "[abc]3", MatchInfo({}, {}));
    test("abba", "!select{M{max}}a#M(b*)a", MatchInfo({}, {1, 3}));
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!dfa;!select{fieldname{ca}fieldbody{ca}}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!dfa;!select{fieldname{ca}fieldbody{ca}}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("LINE\r\nFirst:Second\r\n\r\n",
    "!select{fieldname{ca}}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
    MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}}, {6, 11}));
    test("LINE\r\nFirst:Second\r\n\r\n",
    "!select{fieldname}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
    MatchInfo({{0, 12}, {1, 18}}, {6, 11}));
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!select{fieldname{ca}fieldbody{ca}}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("абвгд", "абвгд", MatchInfo({}, {}));
    test("абвввввввгд", "абв*г+д", MatchInfo({}, {}));
    test("абвввввввд", "абв*г+д", MatchInfo());
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
    "!dfa;LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
    MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {}));
    test("LINE\r\nFirst:Second\r\n\r\n",
        "LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}}, {}));
    test("GET / HTTP/1.1\r\nHost: example.com\r\nAAAAA: a\rfaafafdf\r\n\r\n",
        "!dfa;(GET|POST) / HTTP/(1.1|1.0|0.9)\r\n([\\u0021-\\u007E&^:]+:([\\u0000-\\u007F&^\r\n])*\r\n)*\r\n",
        MatchInfo());
    test("\r24234\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo());
    test("\n3432\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo());
    test("3:::;;432\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo({}, {}));
    test("3:::;;432   \r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo({}, {}));
    test("b", "#boba(b)", MatchInfo({{0, 0}, {1, 1}}, {}));
    test("abc", "!selarr{boba{ca}}a#boba(b)c", MatchInfo({{0, 1}, {1, 2}}, {1, 2}));
    for (int i = 0; i < 64; i++) {
        std::string T;
        T += ('a' + (i >> 3));
        T+= ('a' + (i % 8));
        test(T, "(((a|b)|(c|d))|((e|f)|(g|h)))!r{2}", MatchInfo({}, {}));
    }
    test("abba", "!dfa;!select{M{max}}a#M(b*)a", MatchInfo({}, {1, 3}));
    test("abba", "!select{M{max}}a#M(!any;*)a", MatchInfo({}, {1, 3}));
    test("abba", "!dfa;!select{M{max}}a#M(!any;*)a", MatchInfo({}, {1, 3}));
    test("b3", "[abc]3", MatchInfo({}, {}));
    test("c3", "[abc]3", MatchInfo({}, {}));
    test("aa", "aa", MatchInfo({}, {}));
    test("aaaaa", "a*", MatchInfo({}, {}));
    test("bababbaa", "[ab]*", MatchInfo({}, {}));
    test("bababbaa", "!dfa;[ab]*", MatchInfo({}, {}));
    return 0;
}
