#ifndef LIBREGEXIS024_INSTRUCTION_IMPLEMENTATION_H
#define LIBREGEXIS024_INSTRUCTION_IMPLEMENTATION_H

/* This file should not be included outside libregex024 virtual machine implementation */

#include <libregexis024vm/libregexis024vm.h>
#include <libregexis024vm/vm_opcodes.h>
#include <assert.h>

#define smitsya(error_type) do {ctx.error = error_codes::error_type; return; } while (0)

#define SLOT_EMPTY_val 0
#define SLOT_OCCUPIED 1
#define SLOT_OCCUPIED_val SLOT_OCCUPIED
#define SLOT_NEW 2
#define SLOT_NEW_val (SLOT_OCCUPIED | SLOT_NEW)

#define check_available_prg(regionSz) if (!check_inboundness(ctx, regionSz)){ \
                            ctx.error = error_codes::improper_finish; return; }


#if defined(LIBREGEXIS024_DEBUG) && defined(LIBREGEXIS024_ALLOW_LOUD)
#include <debugging_regexis024/vm/libregexis024vm_debug.h>
#define ctx_print_debug(ctx) debug_print_context(ctx, __func__)
#define thread_print_debug(thread) debug_print_thread(thread, __func__)
#else
#define ctx_print_debug(ctx)
#define thread_print_debug(thread)
#endif

namespace regexis024 {
    bool check_inboundness(VMContext& ctx, int region);

    uint8_t extract_b(VMContext& ctx);
    uint16_t extract_w(VMContext& ctx);
    uint32_t extract_dw(VMContext& ctx);
    uint64_t extract_qw(VMContext& ctx);

    uint8_t extract_instruction(VMContext& ctx);
    sslot_id_t extract_sslot_id(VMContext& ctx);
    near_ptr_t extract_near_pointer(VMContext& ctx);
    tai_t extract_track_array_index(VMContext& ctx);

    void try_to_continue_scheduled(VMContext& ctx);

    void instruction_table(VMContext& ctx);
}

#endif //LIBREGEXIS024_INSTRUCTION_IMPLEMENTATION_H
