#ifndef LIBREGEXIS024_LIBREGEXIS024VM_H
#define LIBREGEXIS024_LIBREGEXIS024VM_H

/* This thing is bloated. And slow (Because I designed it imperfectly and because it is bloated).
* I could have halven the amount of bloat, but that would require me writing code in headers.
* I am gonna use it for KM, even more bloated project. So I thought that this design is on the spot.
* C++ is such a funny language. Code is divided into .cpp and .h files. But it only makes problems.
* All of my work on this C++ project was not serious from the beginning. It's all funny stuff. */

/* Also, please, consider using libregexis024vm/libregexis024vm_interface.h
* Naming in this project is super inconsistent. I don't want it to trash your namespace */

#include <libregexis024vm/vm_errno.h>
#include <libregexis024vm/vm_opcodes_types.h>
#include <stdlib.h>
#include <stdint.h>

namespace regexis024 {
    struct SSID_Stack{
        sslot_id_t* slots = NULL;
        sslot_id_t max_size = 0;
        sslot_id_t sz = 0;

        sslot_id_t pop();
        void append(sslot_id_t x);
        bool empty() const;

        SSID_Stack(const SSID_Stack&) = delete;
        SSID_Stack& operator=(const SSID_Stack&) = delete;
        SSID_Stack() = default;

        ~SSID_Stack();
    };

    struct CollectionArrayNode{
        /* Key is small for historical reasons I do not rememeber. Who cares anyway */
        tai_t key;
        uint64_t value;
        /* NULL at the beginning */
        CollectionArrayNode* prev;
        /* Reference counting */
        uint64_t refs = 0;
    };

    struct Thread{
        /* First byte field is used only when thread is located in slot */
        uint8_t slot_occupation_status = 0;
        near_ptr_t IP = 0;
        CollectionArrayNode* CAHptr = NULL;
        /* Pointer to the seletion array. SA's are reference counted. Because of that every SA
        * is elongated by one meta element in the beginning - reference counter. So the actual elements
        * are enumerated starting from one. */
        uint64_t* SAptr = NULL;

        void delete_thread() noexcept;
    };

    struct VMContext{
        VMContext(size_t programSize, const uint8_t *data, uint64_t caTreeLimit, tai_t saLenLimit,
                            sslot_id_t readSsLimit, sslot_id_t forkSsLimit, uint64_t timeTickLimit);

        error_code_t feedSOF();
        /* You can safely pile up calls to this command, nothing bad will happen */
        error_code_t startThread();
        error_code_t extendedFeedCharacter(uint64_t input);
        error_code_t feedCharacter(uint64_t INP, uint64_t corresponding_byte_amount);


        ~VMContext();

        /* Program size larger than 2^62 is forbidden */
        size_t program_size = 0;
        const uint8_t* prg = NULL;

        /* Max allowed index of CA is 2^16 - 1
        * Max allowed index of SA is 2^16 - 1. VM can be configured to allow even less */
        /* CA = Collecton array. */
        uint64_t CA_TREE_LIMIT;
        /* SA = Selection array */
        tai_t SA_LEN_LIMIT;
        sslot_id_t READ_SS_LIMIT;
        sslot_id_t FORK_SS_LIMIT;

        /* If time_tick_limit is non-zero, regex virtual machine will stop with error
        * after this many ticks. This parameter set's the timeout.*/
        uint64_t time_tick_limit;

        /* This context is used only for one FA match session. This field measures each tick
        * timer <= time_tick_limit */
        uint64_t timer = 0;
        /* CAN_total <= CA_TREE_LIMIT */
        uint64_t CAN_total = 0;

        /* Program selects it */
        tai_t selection_array_len = 0;
        sslot_id_t read_slots_number = 0;
        sslot_id_t fork_slots_number = 0;

        bool have_sift_function = false;
        near_ptr_t sift_function;

        bool initialized = false;
        near_ptr_t unnatural_started_thread_IP = 1337;
        error_code_t error = error_codes::stable;

        Thread* READ_halted_slots;
        SSID_Stack READ_halted_stack_old;
        SSID_Stack READ_halted_stack_new;
        Thread* FORK_halted_slots;
        SSID_Stack FORK_halted_stack;

        Thread active_thread;

        /* Environment for sifting stuff */
        Thread* sifting_with = NULL;
        /* specifies the type of operation vm should do after shift (there are only two distinct options) */
        uint8_t who_started_sift;
        /* Sifting process uses IP field of active thread. Other data of thread is not modified or used during collision
        * procudure. Old IP is stored there, if needed */
        near_ptr_t child_ret_IP;
        near_ptr_t intruder_IP;
        /* RAX corresponds to intruder. Its data is stored in active thread field*/
        uint64_t RAX;
        /* RBX corresponds to homesteader. Its data is accessible by `REGEX_IS024_Thread* sifting_with` pointer*/
        uint64_t RBX;

        /* Will be unoccupied if no threads matched. After each feed of character this field will be wiped
        * User should take care of intermediate success himself */
        Thread matched_thread;

        uint64_t INP = 0;
        uint64_t passed_chars = 0;
        uint64_t passed_bytes = 0;
    };

    void wipe(VMContext& ctx) noexcept;
}
#endif //LIBREGEXIS024_LIBREGEXIS024VM_H
