#ifndef LIBREGEXIS024_LIBREGEXIS024VM_INTERFACE_H
#define LIBREGEXIS024_LIBREGEXIS024VM_INTERFACE_H

#include <stdint.h>
#include <vector>
#include <libregexis024vm/vm_errno.h>
#include <libregexis024vm/vm_opcodes_types.h>

namespace regexis024 {
    struct CAEvent{
        tai_t key;
        uint64_t value;
        bool operator==(const CAEvent& other) const;
    };

    struct VirtualMachine{
        VirtualMachine(size_t programSize, const uint8_t *data, uint64_t caTreeLimit, uint16_t saLenLimit,
                                   uint32_t readSsLimit, uint32_t forkSsLimit, uint64_t timeTickLimit);

        VirtualMachine(const VirtualMachine& ) = delete;
        VirtualMachine& operator=(const VirtualMachine&) = delete;

        error_code_t initialize();
        bool isInitialized();
        bool isUsable();
        virtual ~VirtualMachine();
        tai_t getSelectionArrayLength();
        error_code_t getErrno();
        bool haveSurvivors();
        bool isMatched();
        std::vector<CAEvent> getMatchedThreadCABranchReverse();
        uint64_t getMatchedThreadSAValue(uint16_t key);

        error_code_t addNewMatchingThread();
        error_code_t feedCharacter(uint64_t input, uint64_t bytesResembled);
        void wipeToInit();
    private:
        bool gave_SOF = false;
        void* opaque;
    };
}

#endif //LIBREGEXIS024_LIBREGEXIS024VM_INTERFACE_H
