#ifndef LIBREGEXIS024_VM_ERRNO_H
#define LIBREGEXIS024_VM_ERRNO_H

#include <stdint.h>

namespace regexis024 {
    namespace error_codes {
        enum regex024_error_code_I: int {
            stable = 0,
            ca_tree_limit_violation = -1,
            sa_length_limit_violation = -2,
            read_sslot_count_limit_violation = -3,
            fork_sslot_count_limit_violation = -4,
            timeout = -5,
            /* Threads should be either abandoned by user of virtual machine after MATCH,
             * ot be stopped by DIE instruction. Out of bound jump is disallowed */
            improper_finish = -6,
            /* Operation for general phase is executed in init phase */
            too_early = -7,
            /* Operation for init phase is executed in general phase */
            too_late = -8,
            /* Used selection array index is out of range */
            selection_arr_out_of_range = -9,
            /* Used read slot is out of range */
            read_sslot_out_of_range = -10,
            /* Used fork slot is out of range */
            fork_sslot_out_of_range = -11,

            invalid_opcode = -12,
            invalid_register_code = -13,
            /* Next operation scheduled for execution is forbidden in general thread */
            instruction_not_for_general_thread = -14,
            /* Next operation scheduled for execution is forbidden in collision thread */
            instruction_not_for_collision_thread = -15,
            /* Program willingly threw exception */
            program_throw = -16,
            /* O_o */
            bad_alloc = -17,
        };
    }

    typedef error_codes::regex024_error_code_I error_code_t;

    const char* error_code_to_str(error_code_t x);
}

#endif //LIBREGEXIS024_VM_ERRNO_H
